# MT::ObjectDriver::Redland
# -----------------
# $Revision: 1.21 $
# $Date: 2005/04/08 22:43:09 $
# ---------------------------------------------------------

=head1 NAME

MT::ObjectDriver::Redland - MovableType driver to add a Redland storage mechanism.

=cut

package MT::ObjectDriver::Redland;

=head1 SYNOPSIS

Once installed in your MovableType lib/MT/ObjectDriver directory,
the standard database configuration directives can be used in mt.cfg:

 ObjectDriver Redland
 Database greg
 DBUser greg
 DBHost localhost

The Redland driver currently only uses the Redland mysql storage type. The
database password should be stored in mt-db-pass.cgi, as with any other
database driver.

=head1 REQUIRES

 MovableType 3.1 or higher
 RDF::Redland
 Sort::Naturally
 Encode

=head1 DESCRIPTION

MT-Redland is an attempt to integrate Movable Type with the Redland RDF
Framework. It's goals are to provide a stable database backend for Movable
Type, and a plugin framework for enabling easy semantic annotation of weblog
data.

=cut

use 5.006;

use MT;
use strict;
use warnings;
use Carp qw(carp confess);

use Encode;
use RDF::Redland;
use Data::Dumper;
use Sort::Naturally;
use Tie::Cache::LRU;

use MT::Util qw( offset_time_list );
use base qw(MT::ObjectDriver);

our ($VERSION, $REVISION, $debug, $blankseq);
BEGIN {
	$debug				= 0;
	$blankseq			= 1;
	$REVISION			= do { my @REV = split(/\./, (qw$Revision: 1.21 $)[1]); sprintf("%0.3f", $REV[0] + ($REV[1]/1000)) };
	$VERSION			= 0.007;
}

sub init {
	my $driver = shift;
	$driver->SUPER::init(@_);
	my $cfg		= $driver->cfg;
	my $db		= $cfg->Database;
	my $host	= $cfg->DBHost;
	my $user	= $cfg->DBUser;
	my $pass	= $cfg->DBPassword;
	unless ($pass) {
		my @dirs	= ('.', $cfg->CGIPath, map { substr($_,0,rindex($_,'/')) } map {$cfg->$_()} qw(TemplatePath CSSPath ImportPath PluginPath SearchTemplatePath));
		foreach my $path (@dirs) {
			my $pass_file	= File::Spec->catfile($path, 'mt-db-pass.cgi');
			if (-r $pass_file) {
				if (open my $fh, $pass_file) {
					$pass = <$fh>;
					close($fh);
					if ($pass) {
						chomp($pass);
						$pass =~ s!^\s*!!;
						$pass =~ s!\s*$!!;
						$cfg->DBPassword($pass);
						last;
					}
				}
			}
		}
	}
	my $error;
	RDF::Redland::set_error_handler(sub { $error = shift; });
	my @args	= ("mysql", 'mt', qq(host='localhost',database='${db}',user='${user}',password='${pass}'));
	my $storage	= new RDF::Redland::Storage(@args);
	if ($error) {
		undef $error;
		$args[2]	.= ",new='yes'";
		$storage	= new RDF::Redland::Storage(@args);
	}
	my $model	= new RDF::Redland::Model($storage, "");
	$driver->{model} = $model;
	return $driver->error(MT->translate("Connection error: [_1]", $error)) if ($error);
	$driver;
}

sub model {
	my $driver	= shift;
	return $driver->{model};
}

sub load {
	my $driver = shift;
	carp "load ( @_ )" if ($debug);
	my $iter	= $driver->load_iter( @_ );
	my @data;
	while (my $obj = $iter->()) {
		return $obj unless wantarray;
		push(@data, $obj);
	}
	return @data;
}

sub load_iter {
	my $driver = shift;
	$driver->run_callbacks($_[0] . '::pre_load', \@_);
	carp "load_iter ( @_ )" if ($debug);
	my ($class, $terms, $args) = @_;
	$args		||= {};
	$terms		= {} unless (defined($terms));
	$terms		= { "id" => $terms } if ($terms and not ref($terms));
	my $urin	= $class->get_rdf_class_node;
	my $uri		= $urin->uri->as_string;
	my $model	= $driver->{model};
	
	my $nodes;
	if (my $join = $args->{'join'}) {
		my ($jclass, $jcol, $jterms, $jargs)	= @{ $join };
		my $iter	= $jclass->load_iter( $jterms, $jargs );
		my $pred	= $class->get_column_rdf_node( 'id' );
		my $const	= $class->get_rdf_column_constructor( 'id' );
		$args->{'unique'}	||= $jargs->{'unique'} if (ref($jargs));
		$nodes		= sub {
			my $jnode	= $iter->();
			return undef unless $jnode;
			my $col		= $jnode->$jcol();
			my $val		= RDF::Redland::Node->$const( $col );
			my ($node)	= $model->sources( $pred, $val );
			return $node;
		};
	} else {
		my $i		= $driver->_load_iterator($class, $terms, $args);
		$nodes		= sub {
			return undef if (not $i or not ref($i));
			my $node	= $i->();
			if ($node and exists($args->{'rdql'})) {	# make sure the rdql is returning the right type of node
				my $typen	= $class->get_rdf_type_node;
				my $type	= $typen->uri->as_string;
				my $st		= RDF::Redland::Statement->new( $node, map { RDF::Redland::Node->new_from_uri($_) } ('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $type) );
				$node		= undef unless ($model->contains_statement( $st ));
			}
			return $node;
		};
	}
	
	$nodes			= $driver->_sort_nodes( $nodes, $class, $terms, $args );
	my $stream	= $driver->_make_stream( $nodes, $class, $terms, $args );
	return $stream;
}

{
my %object_cache;
tie %object_cache, 'Tie::Cache::LRU', 100;
sub _make_stream {
	my $driver	= shift;
	my $nodes	= shift;
	my $class	= shift;
	my $terms	= shift;
	my $args	= shift;
	my $model	= $driver->{model};
	
	my %unique;
	my $unique		= $args->{'unique'};
	my $finished	= 0;
	my $limit		= $args->{'limit'};
	my $iter		= sub {
		my $node;
		while (1) {
			return undef if ($finished);
			if ($limit) {
				$limit--;
				$finished	= 1 if ($limit == 0);
			}
			$node	= $nodes->();
			unless ($node) {
				$finished	= 1;
				return undef;
			}
			
			next if ($unique and $unique{ $node->as_string }++);
			last;
		}
		my $obj	= $object_cache{$node->as_string};
		unless ($obj) {
			my %data;
			my @cols		= @{ $class->column_names };
			my %datecols	= map { $_ => 1 } qw(created_on modified_on);
			
			my $arcs		= RDF::Redland::Statement->new( $node, undef, undef );
			my @arcs		= $model->find_statements( $arcs );
			my %arcs		= map { $_->predicate->uri->as_string => $_->object } @arcs;
			
			foreach my $col (@cols) {
				my $pred	= $class->get_column_rdf_node( $col );
				if (my $val = $arcs{ $pred->uri->as_string }) {
					if ($val->is_resource) {
						$data{$col}	= $val->uri->as_string;
					} elsif ($val->is_literal) {
						$data{$col}	= decode('utf8', $val->literal_value);
					} else {
						$data{$col}	= $val->blank_identifier;
					}
					$data{$col}	= $driver->db2ts($data{$col}) if ($datecols{ $col });
				} else {
					$data{$col}	= '';
				}
			}
			$obj	= $class->new;
			$obj->set_values(\%data);
			$driver->run_callbacks($class . '::post_load', $terms, $args, $obj);
			warn Dumper($obj) if ($debug > 1);
			
			$object_cache{$node->as_string}	= $obj;
		}
		return $obj;
	};
	
	return $iter;
}

sub save {
	my $driver	= shift;
	carp "save ( @_ )" if ($debug);
	my $obj		= shift;
    my $class	= ref($obj);
	my $model	= $driver->{model};
	my $node	= $obj->get_rdf_node( $model );
    (my $original, $obj)	= ($obj, $obj->clone());
	$object_cache{$node->as_string}	= $obj;
	
	$driver->run_callbacks($class . '::pre_save', $obj, $original);
	my $result;
	if ($driver->exists($obj)) {
		$result	= $driver->update($obj);
	} else {
		$result	= $driver->insert($obj);
	}
	$original->id($obj->id);
	$original->created_on($obj->created_on);
	$original->modified_on($obj->modified_on);
	$driver->run_callbacks($class . '::post_save', $obj);
	return $result;
}

sub remove {
	my $driver = shift;
	carp "remove ( @_ )" if ($debug);
	my $obj		= shift;
	my $class	= ref($obj);
	my $model	= $driver->{model};
	my $node	= $obj->get_rdf_node( $model );
	delete $object_cache{$node->as_string};
	
	$driver->run_callbacks($class . '::pre_remove', $obj);
	my $id		= $obj->id;
	return $driver->error("No such object.") unless defined($id);
	my $pred	= $class->get_column_rdf_node( 'id' );
	my @nodes	= $model->sources( $pred, RDF::Redland::Node->new_literal($id) );
	foreach my $node (@nodes) {
		my @pat		= (
						RDF::Redland::Statement->new( $node, undef, undef ),
						RDF::Redland::Statement->new( undef, undef, $node )
					);
		foreach my $st (@pat) {
			my (@st)	= $model->find_statements( $st );
			$model->remove_statement($_) for (@st);
		}
	}
	$driver->run_callbacks($class . '::post_remove', $obj);
	1;
}

sub remove_all {
	my $driver = shift;
	carp "remove_all ( @_ )" if ($debug);
	my $proto	= shift;
	my $class	= ref($proto) || $proto;
	$driver->run_callbacks($class . '::pre_remove_all');
	my $model	= $driver->{model};
	my $pred	= RDF::Redland::Node->new_from_uri('http://www.w3.org/1999/02/22-rdf-syntax-ns#type');
	my $type	= $class->get_rdf_type_node;
	my @nodes	= $model->sources( $pred, $type );
	foreach my $node (@nodes) {
		delete $object_cache{$node->as_string};
		my @pat		= (
						RDF::Redland::Statement->new( $node, undef, undef ),
						RDF::Redland::Statement->new( undef, undef, $node )
					);
		foreach my $st (@pat) {
			my (@st)	= $model->find_statements( $st );
			$model->remove_statement($_) for (@st);
		}
	}
	$driver->run_callbacks($class . '::post_remove_all');
	1;
} }

sub insert {
	my $driver	= shift;
	carp "insert ( @_ )" if ($debug);
	my $obj		= shift;
	my $cols	= $obj->column_names;
	unless ($obj->id) {
		my $id = $driver->generate_id($obj);
		$obj->id($id);
	}
	if ($obj->properties->{audit}) {
		my $blog_id	= $obj->blog_id;
		my @ts	= offset_time_list(time, $blog_id);
		my $ts	= sprintf '%04d-%02d-%02dT%02d:%02d:%02dZ', $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
		$obj->created_on($ts) unless $obj->created_on;
		$obj->modified_on($ts);
	}
	
	my $model	= $driver->{model};
	my $node	= $obj->get_rdf_node( $model );
	my $uri		= $obj->get_rdf_type_node;
	my $type	= RDF::Redland::Node->new_from_uri('http://www.w3.org/1999/02/22-rdf-syntax-ns#type');
	my $st		= RDF::Redland::Statement->new( $node, $type, $uri );
	if (not $model->contains_statement( $st )) {
		$model->add_statement( $st );
	}
	
	my $col_defs	= $obj->properties->{column_defs};
	my %datecols	= map { $_ => 1 } qw(created_on modified_on);
	foreach my $col (@{ $cols }) {
		my $val		= $obj->column($col) || '';
		next unless ($val);
		$val		= $driver->ts2db($val, $obj, $col) if ($datecols{ $col });
		my $pred	= $obj->get_column_rdf_node( $col );
		my $method	= $obj->get_rdf_column_constructor( $col );
		my $obj		= RDF::Redland::Node->$method( $val );
		my $st		= RDF::Redland::Statement->new( $node, $pred, $obj );
		$driver->{model}->add_statement( $st );
	}
	1;
}

sub update {
	my $driver	= shift;
	carp "update ( @_ )" if ($debug);
	my $obj		= shift;
	my $cols	= $obj->column_names;
	if ($obj->properties->{audit}) {
		my $blog_id	= $obj->blog_id;
		my @ts	= offset_time_list(time, $blog_id);
		my $ts	= sprintf '%04d-%02d-%02dT%02d:%02d:%02dZ', $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
		$obj->created_on($ts) unless $obj->created_on;
		$obj->modified_on($ts);
	}
	
	my $model	= $driver->{model};
	my $node	= $obj->get_rdf_node( $model );
	my %datecols	= map { $_ => 1 } qw(created_on modified_on);
	foreach my $col (@{ $cols }) {
		my $val		= $obj->column($col) || '';
		$val		= $driver->ts2db($val, $obj, $col) if ($datecols{ $col });
		my $pred	= $obj->get_column_rdf_node( $col );
		my $method	= $obj->get_rdf_column_constructor( $col );
		my $st		= RDF::Redland::Statement->new( $node, $pred, undef );
		my @st		= $model->find_statements( $st );
		my $new		= RDF::Redland::Node->$method($val);
		my $add		= 1;
		foreach my $old (@st) {
			if ($old->object->equals( $new )) {
				$add	= 0;
			} else {
				$model->remove_statement( $old );
			}
		}
		next unless ($val and $add);
		if ($add) {
			$st->object( $new );
			$model->add_statement( $st );
		}
	}
	1;
}

sub count {
	my $driver = shift;
	carp "count ( @_ )" if ($debug);
	my ($class, $terms, $args)	= @_;
	my $count	= 0;
	if ($args and exists $args->{'join'}) {
		my $iter	= $class->load_iter( $terms, $args );
		while ($iter->()) {
			$count++;
		}
	} else {
		my $i		= $driver->_load_iterator($class, $terms, $args);
		while ($i->()) {
			$count++;
		}
	}
	return $count;
}

sub exists {
	my $driver = shift;
	carp "exists ( @_ )" if ($debug);
	my $obj		= shift;
    return unless $obj->id;
	my $model	= $driver->{model};
	my $pred	= $obj->get_column_rdf_node( 'id' );
	my $val		= RDF::Redland::Node->new_literal($obj->id);
	my @nodes	= $model->sources( $pred, $val );
	return scalar(@nodes);
}

sub generate_id {
	my $driver	= shift;
	my $class	= shift;
	if (ref($driver) and $class->can('get_rdf_class_node')) {
		my $node	= $class->get_rdf_class_node;
		my $model	= $driver->{model};
		my $pred	= RDF::Redland::Node->new_from_uri('http://kasei.us/e/ns/mt#seq_next');
		my ($stp)	= RDF::Redland::Statement->new( $node, $pred, undef );
		my ($st)	= $model->find_statements( $stp );
		my $val		= 1;
		if ($st) {
			$model->remove_statement($st);
			$val	= $st->object->literal_value;
		}
		$stp->object( RDF::Redland::Node->new_literal( $val+1 ) );
		$model->add_statement($stp);
		return $val;
	} else {
		our $blankseq;
		return sprintf("%s%04d", time, $blankseq++);
	}
}

sub ts2db {
	my $driver	= shift;
	my $date	= shift;
	my $obj		= shift;
	my $col		= shift;
	my $tz		= 'Z';
	if ($obj and $col ne 'modified_on') {
		if (my $blog = $obj->blog) {
			my $offset	= $blog->server_offset;
			my $partial_hour_offset = 60 * abs($offset - int($offset));
			$tz	= sprintf("%s%02d:%02d", $offset < 0 ? '-' : '+', abs($offset), $partial_hour_offset);
		}
	}
	
	if ($date =~ m#^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$#) {
		$date		= "$1-$2-$3T$4:$5:$6${tz}";
	}
	return $date;
}

sub db2ts {
	my $driver	= shift;
	my $date	= shift;
	$date		=~ s/[^0-9]//g;
	$date		=~ s/^(\d{14}).*$/$1/;
	return $date;
}

sub replace_node {
	my $model		= shift;
	my $node		= shift;
	my $url			= shift;
	my $replacewith	= RDF::Redland::Node->new_from_uri( $url );
	return if ($node->equals( $replacewith ));
	my $sameas		= RDF::Redland::Node->new_from_uri( 'http://www.w3.org/2002/07/owl#sameAs' );
	
	foreach ([$node, $replacewith], [$replacewith, $node]) {
		my $equivst		= RDF::Redland::Statement->new( $_->[0], $sameas, $_->[1] );
		unless ($model->contains_statement( $equivst )) {
			$model->add_statement( $equivst );
		}
	}
	warn "Replacing " . $node->as_string . " with " . $replacewith->as_string . "\n" if ($debug);
	foreach my $data (
						[RDF::Redland::Statement->new( $node, undef, undef ), 'subject'],
						[RDF::Redland::Statement->new( undef, undef, $node ), 'object']
					) {
		my ($stmt, $method)	= @{ $data };
		my (@stmts)	= $model->find_statements( $stmt );
		foreach my $st (@stmts) {
			next if ($st->predicate->as_string eq '[http://www.w3.org/2002/07/owl#sameAs]');	# don't touch equivalences
			my $to	= $st->clone;
			$to->$method( $replacewith );
			if ($model->contains_statement( $to )) {
				warn "Model already contains statement: " . $to->as_string if ($debug);
			} else {
				$model->add_statement( $to );
				$model->remove_statement( $st );
				if ($debug > 1) {
					warn "-- \n";
					warn $st->as_string . "\n";
					warn $to->as_string . "\n";
				}
			}
		}
	}
}

sub _sort_nodes {
	my $driver	= shift;
	my $nodes	= shift;
	my $class	= shift;
	my $terms	= shift;
	my $args	= shift;
	my $model	= $driver->{model};
	my $startval	= $args->{'start_val'};
	my $unique		= $args->{'unique'};
	my $offset		= $args->{'offset'} || 0;
	
	if (exists $args->{'sort'}) {
		my $col		= $args->{'sort'};
		my $pred	= $class->get_column_rdf_node( $col );
		my $asc		= (exists $args->{'direction'}) ? ($args->{'direction'} eq 'ascend') : 1;
		my @nodes;
		while (my $node = $nodes->()) {
			push(@nodes, $node);
		}
		@nodes	= sort { ncmp($a->[1], $b->[1]) }
						map {
								my (@val)	= $model->targets( $_, $pred );
								my $val;
								if (@val) {
									if ($val[0]->is_resource) {
										$val	= $val[0]->uri->as_string;
									} elsif ($val[0]->is_literal) {
										$val	= decode('utf8', $val[0]->literal_value);
									} else {
										$val	= $val[0]->blank_identifier;
									}
								}
								[$_, $val]
							}
							@nodes;
		@nodes	= reverse @nodes unless ($asc);
		if ($startval) {
			@nodes	= grep {
							(ncmp($_->[1], $startval) == 1) .. 1
						}
						@nodes;
		}
		@nodes	= map { $_->[0] } @nodes;
		$nodes	= sub {
			return shift(@nodes);
		};
	}
	
	if ($offset) {
		if ($unique) {
			$nodes->() while ($offset--);
		} else {
			my %seen;
			while (my $node = $nodes->()) {
				next if ($seen{ $node->as_string }++);
				last unless --$offset;
			}
		}
	}
	
	return $nodes;
}
		
sub _load_iterator {
	my $driver	= shift;
	my ($class, $terms, $args)	= @_;
	my @tk		= sort { substr($b,-2,2) eq 'id' } keys %{ $terms };
	my @ak		= keys %{ $args };
	my $model	= $driver->{model};
	
	my $i;
	if (scalar(@tk) == 1 and not $args->{'rdql'}) {
		my $col		= shift(@tk);
		my $val		= RDF::Redland::Node->new_literal( $terms->{ $col } );
		my $pred	= $class->get_column_rdf_node( $col );
		my $iter	= $model->sources_iterator( $pred, $val );
		$i			= sub {
			return undef unless ($iter and not $iter->end);
			my $node	= $iter->current;
			$iter->next;
			return $node;
		};
	} elsif (scalar(@tk) == 2 and not $args->{'rdql'}) {
		my $col		= shift(@tk);
		my $val		= RDF::Redland::Node->new_literal( $terms->{ $col } );
		my $pred	= $class->get_column_rdf_node( $col );
		my $iter	= $model->sources_iterator( $pred, $val );
		my $tcol	= shift(@tk);
		my $tpred	= $class->get_column_rdf_node( $tcol );
		my $tval	= RDF::Redland::Node->new_literal( $terms->{ $tcol } );
		my $st		= RDF::Redland::Statement->new( undef, $tpred, $tval );
		$i			= sub {
			return undef if ($iter->end);
			my $node;
			while (not $iter->end) {
				my $test	= $iter->current;
				$st->subject( $test );
				if ($model->contains_statement($st)) {
					$node	= $test;
					last;
				}
			} continue { $iter->next }
			$iter->next unless ($iter->end);
			return $node;
		};
	} else {
		my $rdql		= exists($args->{'rdql'})
						? $args->{'rdql'}
						: $driver->_prepare_rdql($class, $terms, $args);
		my $query	= RDF::Redland::Query->new($rdql);
		my $iter	= $model->query_execute($query);
		$i			= sub {
			return undef if (not $iter or not ref($iter));
			return undef if ($iter->finished);
			my $node	= $iter->binding_value(0);
			$iter->next_result;
			return $node;
		};
	}
	return $i;
}

sub _prepare_rdql {
	my $driver	= shift;
	my $class	= shift;
	my $terms	= shift;
	my $args	= shift;
	$args		||= {};
	$terms		= {} unless (defined($terms));
	my $typen	= $class->get_rdf_type_node;
	my $type	= $typen->uri->as_string;
	my @where;
	push(@where, "(?node rdf:type <${type}>)");
	my $source	= $class->datasource;
	while (my($col,$val) = each(%{ $terms })) {
		no warnings 'uninitialized';
		$val		= qq("$val");
		my $predn	= $class->get_column_rdf_node( $col );
		my $pred	= $predn->uri->as_string;
		push(@where, qq#(?node <${pred}> ${val})#);
	}
	
	my $where	= join("\n\t", @where);
	my $rdql	= "SELECT\n\t?node\nWHERE\n\t${where}\n";
	warn $rdql if ($debug > 1);
	return $rdql;
}

######################################################################
######################################################################
###  Shove these methods into existing MT::Object classes:

use MT::Plugin;
use Digest::SHA1 qw(sha1_hex);

my $plugin	= new MT::Plugin({	name		=> "Redland RDF Plugin, v1.000",
								description	=> "Adds an RDF storage mechanism" });
######################################################################

package MT::Object;

use strict;
use warnings;

sub get_rdf_type_node {
	&get_rdf_class_node;
}
sub get_rdf_class_node {
	my $class	= shift;
	my $name	= $class->datasource;
	my $uri		= "http://kasei.us/e/ns/mt/${name}";
	return RDF::Redland::Node->new_from_uri( $uri );
}
sub get_rdf_node {
	my $self	= shift;
	return $self->{_rdf_node} if ($self->{_rdf_node});
	my $model	= shift || $self->driver->model;
	if (my $id = $self->id) {
		my $pred	= $self->get_column_rdf_node( 'id' );
		my $val		= RDF::Redland::Node->new_literal($id);
		my (@nodes)	= $model->sources( $pred, $val );
		return $nodes[0] if (@nodes);
	}
	my $node	= RDF::Redland::Node->new_from_blank_identifier( MT::ObjectDriver::Redland->generate_id( $self ) );
	$self->{_rdf_node}	= $node;
	return $node;
}
sub get_column_rdf_node {
	my $self	= shift;
	my $col		= shift;
	my %map		= $self->get_column_rdf_node_map;
	my $class	= $self->get_rdf_class_node->uri->as_string;
	my $uri		= (exists $map{ $col }) ? $map{ $col } : join('#', $class, $col);
	return RDF::Redland::Node->new_from_uri( $uri );
}
sub get_column_rdf_node_map {
	return ();
}
sub get_rdf_column_constructor {
	my $self	= shift;
	my $col		= shift;
	my %map		= (
					url		=> 'new_from_uri',
				);
	return $map{ $col } || 'new_literal';
}

######################################################################

package MT::Blog;

use strict;
use warnings;

sub get_rdf_column_constructor {
	my $self	= shift;
	my $col		= shift;
	my %map		= (
					site_url	=> 'new_from_uri',
					archive_url	=> 'new_from_uri'
				);
	return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
}

######################################################################

package MT::Entry;

use strict;
use warnings;

sub get_rdf_type_node {
	return RDF::Redland::Node->new_from_uri('http://purl.org/rss/1.0/item')
}
sub get_rdf_node {
	my $self	= shift;
	my $model	= $self->driver->model;
	my $url		= $self->archive_url;
	if (my $id = $self->id) {
		my $pred	= $self->get_column_rdf_node( 'id' );
		my $val		= RDF::Redland::Node->new_literal($id);
		my ($node)	= $model->sources( $pred, $val );
		if ($node) {
			if ($url ne $node->uri->as_string) {
				MT::ObjectDriver::Redland::replace_node( $model, $node, $url );
			}
			return $node;
		}
	}
	my $node	= RDF::Redland::Node->new_from_uri( $url );
	return $node;
}
sub get_column_rdf_node_map {
	return (
			title		=> 'http://purl.org/rss/1.0/title',
			created_on	=> 'http://purl.org/dc/elements/1.1/date',
			modified_on	=> 'http://purl.org/dc/terms/modified'
		);
}

MT::Entry->add_callback("pre_save", 8, $plugin, sub {
	my ($cb, $obj)	= @_;
	my $model	= $obj->driver->model;
	my $text	= $obj->text || '';
	my $more	= $obj->text_more || '';
	my $desc	= $text . $more;
	my $node	= $obj->get_rdf_node;
	my $l_desc	= RDF::Redland::Node->new_literal( $desc );
	my $l_link	= RDF::Redland::Node->new_literal( $node->uri->as_string );
	my $p_desc	= RDF::Redland::Node->new_from_uri( 'http://purl.org/rss/1.0/description' );
	my $p_link	= RDF::Redland::Node->new_from_uri( 'http://purl.org/rss/1.0/link' );

	my $st		= RDF::Redland::Statement->new( $node, undef, undef );
	foreach ([$p_desc, $l_desc], [$p_link, $l_link]) {
		my ($p, $l)	= @{ $_ };
		$st->predicate( $p );
		my (@st)	= $model->find_statements( $st );
		$model->remove_statement($_) for (@st);
		$st->object( $l );
		$model->add_statement( $st );
	}
});

MT::Entry->add_callback("post_remove", 3, $plugin, sub {
	my ($cb, $obj)	= @_;
	my $model	= $obj->driver->model;
	my $node	= $obj->get_rdf_node;
	my $p_desc	= RDF::Redland::Node->new_from_uri( 'http://purl.org/rss/1.0/description' );
	my $p_link	= RDF::Redland::Node->new_from_uri( 'http://purl.org/rss/1.0/link' );

	my $st		= RDF::Redland::Statement->new( $node, undef, undef );
	foreach my $p ($p_desc, $p_link) {
		$st->predicate( $p );
		my (@st)	= $model->find_statements( $st );
		$model->remove_statement($_) for (@st);
	}
});

######################################################################

package MT::TBPing;

use strict;
use warnings;

sub get_rdf_column_constructor {
	my $self	= shift;
	my $col		= shift;
	my %map		= (
					source_url		=> 'new_from_uri',
				);
	return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
}

######################################################################

package MT::Comment;

use strict;
use warnings;
use Digest::SHA1 qw(sha1_hex);

sub get_rdf_type_node {
	return RDF::Redland::Node->new_from_uri('http://purl.org/net/rssmodules/blogcomments/comment')
}
sub get_column_rdf_node_map {
	return (
			created_on	=> 'http://purl.org/dc/elements/1.1/date',
			modified_on	=> 'http://purl.org/dc/terms/modified',
			text		=> 'http://purl.org/net/rssmodules/blogcomments/body'
		);
}
sub get_rdf_column_constructor {
	my $self	= shift;
	my $col		= shift;
	if ($col eq 'email' and $self->email =~ /^[0-9a-f]{40}$/i) {	# sha1sum
		return 'new_literal';
	} else {
		my %map		= (
						email		=> 'new_from_uri',
						url			=> 'new_from_uri'
					);
		return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
	}
}

MT::Comment->add_callback("pre_save", 8, $plugin, sub {
	my ($cb, $obj)	= @_;
	my $model	= $obj->driver->model;
	my $entry	= MT::Entry->load( $obj->entry_id );
	
	my $name	= $obj->author;
	my $email	= $obj->email;
	my $url		= $obj->url;
	
	my $r_url	= ($url) ? RDF::Redland::Node->new_from_uri( $url ) : undef;
	my $l_name	= ($name) ? RDF::Redland::Node->new_literal( $name ) : undef;
	my $l_mbox;
	if ($email) {
		my $sha;
		if ($email =~ /^[0-9a-f]{40}$/i) {
			$sha	= $email;
		} else {
			$obj->email( "mailto:$email" );
			$sha	= sha1_hex( "mailto:$email" );
		}
		$l_mbox	= RDF::Redland::Node->new_literal( $sha );
	}
	
	my $p_name		= RDF::Redland::Node->new_from_uri( 'http://xmlns.com/foaf/0.1/nick' );
	my $p_url		= RDF::Redland::Node->new_from_uri( 'http://xmlns.com/foaf/0.1/homepage' );
	my $p_mbox		= RDF::Redland::Node->new_from_uri( 'http://xmlns.com/foaf/0.1/mbox_sha1sum' );
	my $p_maker		= RDF::Redland::Node->new_from_uri( 'http://xmlns.com/foaf/0.1/maker' );
	my $p_type		= RDF::Redland::Node->new_from_uri( 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type' );
	my $p_comments	= RDF::Redland::Node->new_from_uri( 'http://purl.org/net/rssmodules/blogcomments/comments' );
	my $t_person	= RDF::Redland::Node->new_from_uri( 'http://xmlns.com/foaf/0.1/Person' );

	my $author;
	if ($l_mbox) {
		($author)	= $model->sources( $p_mbox, $l_mbox );
	}
	
	if ($r_url and not $author) {
		($author)	= $model->sources( $p_url, $r_url );
	}
	
	unless ($author) {
		$author	= RDF::Redland::Node->new_from_blank_identifier('r' . time . 'r' . $MT::ObjectDriver::Redland::blankseq++);
		my $st	= RDF::Redland::Statement->new( $author, $p_type, $t_person );
		$model->add_statement( $st );
	}
	foreach (grep {$_->[1]} [$p_name, $l_name], [$p_url, $r_url], [$p_mbox, $l_mbox]) {
		my ($p, $l)	= @{ $_ };
		my $st	= RDF::Redland::Statement->new( $author, $p, $l );
		$model->add_statement( $st ) unless ($model->contains_statement( $st ));
	}
	
	my $node	= $obj->get_rdf_node;
	$model->add_statement( RDF::Redland::Statement->new( $node, $p_maker, $author ) );
	$model->add_statement( RDF::Redland::Statement->new( $entry->get_rdf_node, $p_comments, $node ) );
});

MT::Comment->add_callback("post_load", 1, $plugin, sub {
	my ($cb, $terms, $args, $obj) = @_;
	my $email	= $obj->email;
	if ($email) {
		$email		=~ s/^mailto://;
		$obj->email( $email );
	}
});

MT::Comment->add_callback("post_remove", 3, $plugin, sub {
	my ($cb, $obj)	= @_;
	my $model	= $obj->driver->model;
	my $node	= $obj->get_rdf_node;
	my $p_maker	= RDF::Redland::Node->new_from_uri( 'http://xmlns.com/foaf/0.1/maker' );
	
	my $st		= RDF::Redland::Statement->new( $node, $p_maker, undef );
	my @st		= $model->find_statements( $st );
	$model->remove_statements( $_ ) for (@st);
});

######################################################################

package MT::Author;

use strict;
use warnings;

sub get_rdf_type_node {
	return RDF::Redland::Node->new_from_uri('http://xmlns.com/foaf/0.1/Person')
}
sub get_column_rdf_node_map {
	return (
			name		=> 'http://xmlns.com/foaf/0.1/name',
			nickname	=> 'http://xmlns.com/foaf/0.1/nick',
			email		=> 'http://xmlns.com/foaf/0.1/mbox',
			mboxsha		=> 'http://xmlns.com/foaf/0.1/mbox_sha1sum',
			url			=> 'http://xmlns.com/foaf/0.1/homepage'
		);
}
sub get_rdf_column_constructor {
	my $self	= shift;
	my $col		= shift;
	my %map		= (
					email		=> 'new_from_uri',
					url			=> 'new_from_uri'
				);
	return $map{ $col } || $self->SUPER::get_rdf_column_constructor( $col );
}
sub column_names {
	my $self	= shift;
	my $names	= $self->SUPER::column_names;
	push(@{ $names }, qw(mboxsha));
	return $names;
}

MT::Author->add_callback("post_load", 1, $plugin, sub {
	my ($cb, $terms, $args, $obj) = @_;
	my $email	= $obj->email;
	$email		=~ s/^mailto://;
	$obj->email( $email );
});

MT::Author->add_callback("pre_save", 1, $plugin, sub {
	my ($cb, $obj)	= @_;
	if (my $email = $obj->email) {
		$email	= "mailto:${email}";
		$obj->email( $email );
		my $sha	= sha1_hex( $email );
		$obj->mboxsha( $sha );
	} else {
		delete $obj->{'column_values'}{'email'};
	}
});

######################################################################

MT::Placement->add_callback("pre_save", 3, $plugin, sub {
	my ($cb, $obj)	= @_;
	my $model		= $obj->driver->model;
	my $entry		= MT::Entry->load( $obj->entry_id );
	my $cat			= MT::Category->load( $obj->category_id );
	my $enode		= $entry->get_rdf_node;
	my $cnode		= $cat->get_rdf_node;
	if ($enode and $cnode) {
		my $p_subj		= RDF::Redland::Node->new_from_uri('http://purl.org/dc/elements/1.1/subject');
		my ($l_subj)	= $model->targets( $cnode, MT::Category->get_column_rdf_node( 'label' ) );
		my $st	= RDF::Redland::Statement->new( $enode, $p_subj, $l_subj );
		$model->add_statement( $st );
	}
});

MT::Placement->add_callback("post_remove", 3, $plugin, sub {
	my ($cb, $obj)	= @_;
	my $model		= $obj->driver->model;
	my $entry		= MT::Entry->load( $obj->entry_id );
	my $cat			= MT::Category->load( $obj->category_id );
	my $enode		= $entry->get_rdf_node;
	my $cnode		= $cat->get_rdf_node;
	if ($enode and $cnode) {
		my $p_subj		= RDF::Redland::Node->new_from_uri('http://purl.org/dc/elements/1.1/subject');
		my ($l_subj)	= $model->targets( $cnode, MT::Category->get_column_rdf_node( 'label' ) );
		my $st	= RDF::Redland::Statement->new( $enode, $p_subj, $l_subj );
		my @st	= $model->find_statements( $st );
		$model->remove_statement( $_ ) for (@st);
	}
});

1;

__END__

=head1 REVISION HISTORY

 $Log: Redland.pm,v $
 Revision 1.21  2005/04/08 22:43:09  greg
 - bumped version to 0.007
 - updated README with changelog
 - updated META.yml

 Revision 1.20  2005/04/08 22:33:43  greg
 - Placements now don't delete other dc:subjects on save (was breaking multiple categories)
 - _load_iterator now uses RDQL only for loads with more than two terms to match (major speed improvement)
 - _load_iterator now returns a closure (not a redland Iterator)

 Revision 1.19  2005/04/08 20:07:00  greg
 - add checks in load_iter to verify node type if loading is based on unverified rdql
 - fix caching code to update/remove cached objects in save() and remove()

 Revision 1.18  2005/04/07 02:39:02  greg
 - added description to POD
 - bumped version to 0.006

 Revision 1.17  2005/04/07 02:13:41  greg
 - added caching code using Time::Cache::LRU (not sure if this will pay off)
 - broke code off from load_iter as _sort_nodes and _make_stream
 - object construction now loads all data from redland with one find_statements call
 - remove code now looks for statements where the object is the target of the triple
 - comment email field now can change to a literal if the value is a sha1sum
 - fixed predicate for attaching comments to entries (forgot to change after copy-paste)
 - if there's an existing author for a new comment, we add new triples (not deleting old values)

 Revision 1.16  2005/04/06 22:24:48  greg
 - when creating a new mt storage model, avoids creating a model object the
   first time around while the storage object is going to be undef.
 - now avoids adding triples if the object is ''
 - object loading now does fixup when the data isn't in a triple (sets value to '')
 - rdf node is now cached in MT::Object objects
   (was breaking with multiple calls to get_rdf_node before it had been saved to the database)
 - added use strict/warnings to all the MT::Object subclass packages
 - added callback code for Comments to attach them to entries and authors (adds foaf:Person data)

 Revision 1.15  2005/04/06 19:15:29  greg
 - removed generation of sha1sum in Comment callbacks (need to get the semantics right on these)
 - fixed type in POD

 Revision 1.14  2005/04/06 15:09:45  greg
 - bumped version number to 0.005
 - updated email address in AUTHOR POD section

 Revision 1.13  2005/04/05 20:49:55  greg
 - added rdf mappings and callbacks for Comments
 - added remove callback for removing Entry link/description data

 Revision 1.12  2005/04/05 18:30:44  greg
 - if model construction fails, now tries to create it with "new='yes'" arg
 - added callbacks for MT::Placement to add dc:subject triples for categories

 Revision 1.11  2005/04/05 11:41:12  greg
 - bumped version number
 - removed unused modules

 Revision 1.10  2005/04/05 11:26:36  greg
 - pulled redland iterator generation from load_iter and count to new _load_iterator
 - added support for offset and start_val in load and load_iter %args

 Revision 1.9  2005/04/05 05:33:47  greg
 - now reads password file and sets $cfg->DBPassword to avoid having to patch MT
 - uses Redland error handler to display error when creating the model fails

 Revision 1.8  2005/04/04 10:48:29  greg
 - bumped version number to 0.003
 - moved sorting in load_iter to happen before object construction (now works on Redland Nodes)
 - implemented more efficient count method (instead of calling load_iter)

 Revision 1.7  2005/04/04 07:51:19  greg
 - now properly removes old rss:link and rss:description triples on update

 Revision 1.6  2005/04/03 23:54:39  greg
 - updated entry predicates to use rss terms
 - added Entry pre_save callback to generate rss link and description triples

 Revision 1.5  2005/04/03 22:16:21  greg
 - updated VERSION number

 Revision 1.4  2005/04/02 22:51:12  greg
 - updated the doc_link MT plugin variable to point to MT-Redland
 - CMS hooks can now add multiple boxes instead of all being in the same box

 Revision 1.3  2005/04/02 22:37:58  greg
 - implemented join, unique, limit, and sort in load_iter()

 Revision 1.2  2005/04/01 21:19:51  greg
 - Updating an entry with a new URL now does the Right Thing
   (including inserting owl:sameAs triples)


=head1 COPYRIGHT

Copyright (c) 2005 Gregory Williams. All rights reserved. This program is free
software; you can redistribute it and/or modify it under the same terms as Perl
itself.

=head1 AUTHOR

Gregory Williams <greg@evilfunhouse.com>

=cut
